"""Vizio SmartCast device zeroconf discovery function and classes."""

import time
from typing import Callable, List

from pyvizio.const import DEFAULT_TIMEOUT
from zeroconf import IPVersion, ServiceBrowser, ServiceInfo, Zeroconf


class ZeroconfDevice(object):
    def __init__(self, name: str, ip: str, port: int, model: str, id: str) -> None:
        self.name = name
        self.ip = ip
        self.port = port
        self.model = model
        self.id = id

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


class ZeroconfListener(object):
    """Basic zeroconf listener."""

    def __init__(self, func: Callable[[ServiceInfo], None]) -> None:
        """Initialize zeroconf listener with function callback."""
        self._func = func

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__

    def add_service(self, zeroconf: Zeroconf, type: str, name: str) -> None:
        """Callback function when zeroconf service is discovered."""
        self._func(zeroconf.get_service_info(type, name))


def discover(service_type: str, timeout: int = DEFAULT_TIMEOUT) -> List[ZeroconfDevice]:
    """Return all discovered zeroconf services of a given service type over given timeout period."""
    services = []

    def append_service(info: ServiceInfo) -> None:
        """Append discovered zeroconf service to service list."""
        name = info.name[: -(len(info.type) + 1)]
        ip = info.parsed_addresses(IPVersion.V4Only)[0]
        port = info.port
        model = info.properties.get(b"name", "").decode("utf-8")
        id = info.properties.get(b"id")

        # handle id decode for various discovered use cases
        if isinstance(id, bytes):
            try:
                int(id, 16)
            except Exception:
                id = id.hex()
        else:
            id = None

        service = ZeroconfDevice(name, ip, port, model, id)
        services.append(service)

    zeroconf = Zeroconf()
    ServiceBrowser(zeroconf, service_type, ZeroconfListener(append_service))
    time.sleep(timeout)
    zeroconf.close()

    return services



"""pyvizio constants."""

DEVICE_CLASS_SOUNDBAR = "soundbar"  # Deprecated
DEVICE_CLASS_SPEAKER = "speaker"
DEVICE_CLASS_TV = "tv"

DEFAULT_DEVICE_ID = "pyvizio"
DEFAULT_DEVICE_CLASS = DEVICE_CLASS_TV
DEFAULT_DEVICE_NAME = "Python Vizio"
DEFAULT_PORTS = [7345, 9000]
DEFAULT_TIMEOUT = 5

MAX_VOLUME = {DEVICE_CLASS_TV: 100, DEVICE_CLASS_SPEAKER: 31}

# Current Input when app is active
INPUT_APP = "SMARTCAST"

# App name returned when it is not in app dictionary
UNKNOWN_APP = "_UNKNOWN_APP"
NO_APP_RUNNING = "_NO_APP_RUNNING"



"""pyvizio utility module."""

import json
from typing import Any, Dict, List, Tuple

from pyvizio.util.const import (
    APK_SOURCE_PATH,
    APP_NAMES_FILE,
    APP_PAYLOADS_FILE,
    RESOURCE_PATH,
)


def gen_apps_list_from_src(
    apk_source_path: str = APK_SOURCE_PATH, resource_path: str = RESOURCE_PATH
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """Parse JSON from VizioCast Android app source in `apk_source_path`/`resource_path` and return list of apps for use in pyvizio."""
    base_path = f"{apk_source_path}/{resource_path}"
    app_names_filepath = f"{base_path}/{APP_NAMES_FILE}"
    app_configs_filepath = f"{base_path}/{APP_PAYLOADS_FILE}"

    with open(app_names_filepath) as f:
        app_names = json.load(f)

    with open(app_configs_filepath) as f:
        app_configs = json.load(f)

    pyvizio_apps = []

    for app_name in app_names:
        # returns first app where condition is true
        app_config = next(
            (
                app_config
                for app_config in app_configs
                if app_config["id"] == app_name["id"]
            )
        )

        if app_config:
            config_json = app_config["chipsets"]["*"][0]["app_type_payload"]
            config = json.loads(config_json)
            pyvizio_apps.append(
                {
                    "name": app_name["name"],
                    "country": [country.lower() for country in app_name["country"]],
                    "id": app_name["id"],
                    "config": config,
                }
            )

    return pyvizio_apps



"""Vizio SmartCast API commands for audio settings."""

from typing import Any, Dict, Union

from pyvizio._api._protocol import (
    ENDPOINT,
    TYPE_EQ_LIST,
    TYPE_EQ_SLIDER,
    TYPE_VALUE,
    ResponseKey,
)
from pyvizio._api.item import Item, ItemCommandBase, ItemInfoCommandBase
from pyvizio.helpers import dict_get_case_insensitive


class GetAllAudioSettingsCommand(ItemInfoCommandBase):
    """Command to get list of all audio setting names and corresponding values."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get list of all audio setting names and corresponding values."""
        super(GetAllAudioSettingsCommand, self).__init__(device_type, "AUDIO_SETTINGS")

    def process_response(self, json_obj: Dict[str, Any]) -> Dict[str, Union[int, str]]:
        """Return response to command to get list of all audio setting names and corresponding values."""
        items = [
            Item(item)
            for item in dict_get_case_insensitive(json_obj, ResponseKey.ITEMS, [])
        ]
        audio_settings = {}
        for item in items:
            if item.type.lower() in (TYPE_EQ_LIST, TYPE_EQ_SLIDER, TYPE_VALUE):
                audio_settings[item.c_name] = item.value
        return audio_settings


class GetAudioSettingCommand(ItemInfoCommandBase):
    """Command to get value of an audio setting by name."""

    def __init__(self, device_type: str, setting_name: str) -> None:
        """Initialize command to get value of an audio setting by name."""
        super(GetAudioSettingCommand, self).__init__(device_type, "AUDIO_SETTINGS", 0)
        self.item_name = setting_name.lower()
        ItemInfoCommandBase.url.fset(
            self, f"{ENDPOINT[device_type]['AUDIO_SETTINGS']}/{setting_name}"
        )


class ChangeAudioSettingCommand(ItemCommandBase):
    """Command to set value of an audio setting by name to new value."""

    def __init__(
        self, device_type: str, id: int, setting_name: str, value: Union[int, str]
    ) -> None:
        """Initialize command to set value of an audio setting by name to new value."""
        super(ChangeAudioSettingCommand, self).__init__(
            device_type, "AUDIO_SETTINGS", id, value
        )
        ItemCommandBase.url.fset(
            self, f"{ENDPOINT[device_type]['AUDIO_SETTINGS']}/{setting_name}"
        )



"""Vizio SmartCast device SSDP discovery function and classes."""

#   Copyright 2014 Dan Krause
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import http.client
import io
import socket

from pyvizio.const import DEFAULT_TIMEOUT


class SSDPDevice(object):
    """Representation of Vizio device discovered via SSDP."""

    def __init__(self, ip, name, model, udn) -> None:
        self.ip = ip
        self.name = name
        self.model = model
        self.udn = udn

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


class SSDPResponse(object):
    """SSDP discovery response."""

    class _FakeSocket(io.BytesIO):
        """Fake socket to retrieve SSDP response."""

        def makefile(self, *args, **kw):
            return self

    def __init__(self, response):
        """Initialize SSDP response."""
        r = http.client.HTTPResponse(self._FakeSocket(response))
        r.begin()
        self.location = r.getheader("location")
        self.usn = r.getheader("usn")
        self.st = r.getheader("st")
        self.cache = r.getheader("cache-control").split("=")[1]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


def discover(service, timeout=DEFAULT_TIMEOUT, retries=1, mx=3):
    """Return all discovered SSDP services of a given service name over given timeout period."""
    group = ("239.255.255.250", 1900)
    message = "\r\n".join(
        [
            "M-SEARCH * HTTP/1.1",
            "HOST: {0}:{1}",
            'MAN: "ssdp:discover"',
            "ST: {st}",
            "MX: {mx}",
            "",
            "",
        ]
    )
    socket.setdefaulttimeout(timeout)
    responses = {}
    for _ in range(retries):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        message_bytes = message.format(*group, st=service, mx=mx).encode("utf-8")
        sock.sendto(message_bytes, group)

        while True:
            try:
                response = SSDPResponse(sock.recv(1024))
                responses[response.location] = response
            except socket.timeout:
                break

        return list(responses.values())



"""Vizio SmartCast API commands and constants for apps."""

from typing import Any, Dict

from pyvizio._api._protocol import ENDPOINT, ResponseKey
from pyvizio._api.base import CommandBase
from pyvizio._api.input import ItemInfoCommandBase
from pyvizio.const import NO_APP_RUNNING, UNKNOWN_APP
from pyvizio.helpers import dict_get_case_insensitive


class AppConfig(object):
    """Vizio SmartCast app config."""

    def __init__(
        self, APP_ID: str = None, NAME_SPACE: int = None, MESSAGE: str = None
    ) -> None:
        self.APP_ID = APP_ID
        self.NAME_SPACE = NAME_SPACE
        self.MESSAGE = MESSAGE

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


class LaunchAppConfigCommand(CommandBase):
    """Command to launch app by config."""

    def __init__(
        self, device_type: str, APP_ID: str, NAME_SPACE: int, MESSAGE: str = None
    ) -> None:
        """Initialize command to launch app by config."""
        super(LaunchAppConfigCommand, self).__init__(
            ENDPOINT[device_type]["LAUNCH_APP"]
        )

        self.VALUE = AppConfig(APP_ID, NAME_SPACE, MESSAGE)


class LaunchAppNameCommand(LaunchAppConfigCommand):
    """Command to launch app by name."""

    def __init__(self, device_type: str, app_name: str) -> None:
        """Initialize command to launch app by name."""
        app_def = next(
            (
                app_def
                for app_def in (APP_HOME + APPS)
                if app_def["name"].lower() == app_name.lower()
            ),
            dict(),
        )

        # Unpack config dict into expected key/value argument pairs
        super(LaunchAppNameCommand, self).__init__(
            device_type, **app_def.get("config", dict())
        )


class GetCurrentAppConfigCommand(ItemInfoCommandBase):
    """Command to get currently running app's config."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get currently running app's config."""
        super(GetCurrentAppConfigCommand, self).__init__(device_type, "CURRENT_APP")

    def process_response(self, json_obj: Dict[str, Any]) -> AppConfig:
        """Return response to command to get currently running app's config."""
        item = dict_get_case_insensitive(json_obj, ResponseKey.ITEM, {})
        current_app_id = dict_get_case_insensitive(item, ResponseKey.VALUE)

        if current_app_id:
            return AppConfig(**current_app_id)

        return AppConfig()


class GetCurrentAppNameCommand(GetCurrentAppConfigCommand):
    """Command to get currently running app's name."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get currently running app's name."""
        super(GetCurrentAppNameCommand, self).__init__(device_type)

    def process_response(self, json_obj: Dict[str, Any]) -> str:
        """Return response to command to get currently running app's name. Returns NO_APP_RUNNING if no app is currently running and UNKNOWN_APP if app name can't be retrieved from APPS."""
        current_app_config = super(GetCurrentAppNameCommand, self).process_response(
            json_obj
        )

        if current_app_config != AppConfig():
            app_def = next(
                (
                    app_def
                    for app_def in (APP_HOME + APPS)
                    if app_def["config"]["APP_ID"] == current_app_config.APP_ID
                    and app_def["config"]["NAME_SPACE"] == current_app_config.NAME_SPACE
                ),
                dict(),
            )

            # Return name of app or UNKNOWN_APP if app name can't be found for given config
            return app_def.get("name", UNKNOWN_APP)

        # Return NO_APP_RUNNING if value from response was "null"
        return NO_APP_RUNNING


APP_HOME = [
    {
        "name": "SmartCast Home",
        "country": ["*"],
        "config": {
            "NAME_SPACE": 4,
            "APP_ID": "1",
            "MESSAGE": "http://127.0.0.1:12345/scfs/sctv/main.html",
        },
    }
]

APPS = [
    {
        "name": "Prime Video",
        "country": ["*"],
        "id": "33",
        "config": {"NAME_SPACE": 2, "APP_ID": "4", "MESSAGE": None},
    },
    {
        "name": "CBS All Access",
        "country": ["usa"],
        "id": "9",
        "config": {"NAME_SPACE": 2, "APP_ID": "37", "MESSAGE": None},
    },
    {
        "name": "CBS News",
        "country": ["usa", "can"],
        "id": "56",
        "config": {"NAME_SPACE": 2, "APP_ID": "42", "MESSAGE": None},
    },
    {
        "name": "Crackle",
        "country": ["usa"],
        "id": "8",
        "config": {"NAME_SPACE": 2, "APP_ID": "5", "MESSAGE": None},
    },
    {
        "name": "Curiosity Stream",
        "country": ["usa", "can"],
        "id": "37",
        "config": {"NAME_SPACE": 2, "APP_ID": "12", "MESSAGE": None},
    },
    {
        "name": "Fandango Now",
        "country": ["usa"],
        "id": "24",
        "config": {"NAME_SPACE": 2, "APP_ID": "7", "MESSAGE": None},
    },
    {
        "name": "FilmRise",
        "country": ["usa"],
        "id": "47",
        "config": {"NAME_SPACE": 2, "APP_ID": "24", "MESSAGE": None},
    },
    {
        "name": "Flixfling",
        "country": ["*"],
        "id": "49",
        "config": {"NAME_SPACE": 2, "APP_ID": "36", "MESSAGE": None},
    },
    {
        "name": "Haystack TV",
        "country": ["usa", "can"],
        "id": "35",
        "config": {
            "NAME_SPACE": 0,
            "APP_ID": "898AF734",
            "MESSAGE": '{"CAST_NAMESPACE":"urn:x-cast:com.google.cast.media","CAST_MESSAGE":{"type":"LOAD","media":{},"autoplay":true,"currentTime":0,"customData":{"platform":"sctv"}}}',
        },
    },
    {
        "name": "Hulu",
        "country": ["usa"],
        "id": "19",
        "config": {"NAME_SPACE": 2, "APP_ID": "3", "MESSAGE": None},
    },
    {
        "name": "iHeartRadio",
        "country": ["usa"],
        "id": "11",
        "config": {"NAME_SPACE": 2, "APP_ID": "6", "MESSAGE": None},
    },
    {
        "name": "NBC",
        "country": ["usa"],
        "id": "43",
        "config": {"NAME_SPACE": 2, "APP_ID": "10", "MESSAGE": None},
    },
    {
        "name": "Netflix",
        "country": ["*"],
        "id": "34",
        "config": {"NAME_SPACE": 3, "APP_ID": "1", "MESSAGE": None},
    },
    {
        "name": "Plex",
        "country": ["usa", "can"],
        "id": "40",
        "config": {"NAME_SPACE": 2, "APP_ID": "9", "MESSAGE": None},
    },
    {
        "name": "Pluto TV",
        "country": ["usa"],
        "id": "12",
        "config": {
            "NAME_SPACE": 0,
            "APP_ID": "E6F74C01",
            "MESSAGE": '{"CAST_NAMESPACE":"urn:x-cast:tv.pluto","CAST_MESSAGE":{"command":"initializePlayback","channel":"","episode":"","time":0}}',
        },
    },
    {
        "name": "RedBox",
        "country": ["usa"],
        "id": "55",
        "config": {"NAME_SPACE": 2, "APP_ID": "41", "MESSAGE": None},
    },
    {
        "name": "TasteIt",
        "country": ["*"],
        "id": "52",
        "config": {"NAME_SPACE": 2, "APP_ID": "26", "MESSAGE": None},
    },
    {
        "name": "Toon Goggles",
        "country": ["usa", "can"],
        "id": "46",
        "config": {"NAME_SPACE": 2, "APP_ID": "21", "MESSAGE": None},
    },
    {
        "name": "Vudu",
        "country": ["usa"],
        "id": "6",
        "config": {
            "NAME_SPACE": 2,
            "APP_ID": "21",
            "MESSAGE": "https://my.vudu.com/castReceiver/index.html?launch-source=app-icon",
        },
    },
    {
        "name": "XUMO",
        "country": ["usa"],
        "id": "27",
        "config": {
            "NAME_SPACE": 0,
            "APP_ID": "36E1EA1F",
            "MESSAGE": '{"CAST_NAMESPACE":"urn:x-cast:com.google.cast.media","CAST_MESSAGE":{"type":"LOAD","media":{},"autoplay":true,"currentTime":0,"customData":{}}}',
        },
    },
    {
        "name": "YouTubeTV",
        "country": ["usa", "mexico"],
        "id": "45",
        "config": {"NAME_SPACE": 5, "APP_ID": "3", "MESSAGE": None},
    },
    {
        "name": "YouTube",
        "country": ["*"],
        "id": "44",
        "config": {"NAME_SPACE": 5, "APP_ID": "1", "MESSAGE": None},
    },
    {
        "name": "Baeble",
        "country": ["usa"],
        "id": "39",
        "config": {"NAME_SPACE": 2, "APP_ID": "11", "MESSAGE": None},
    },
    {
        "name": "DAZN",
        "country": ["usa", "can"],
        "id": "57",
        "config": {"NAME_SPACE": 2, "APP_ID": "34", "MESSAGE": None},
    },
    {
        "name": "FitFusion by Jillian Michaels",
        "country": ["usa", "can"],
        "id": "54",
        "config": {"NAME_SPACE": 2, "APP_ID": "39", "MESSAGE": None},
    },
    {
        "name": "Newsy",
        "country": ["usa", "can"],
        "id": "38",
        "config": {"NAME_SPACE": 2, "APP_ID": "15", "MESSAGE": None},
    },
    {
        "name": "Cocoro TV",
        "country": ["usa", "can"],
        "id": "63",
        "config": {"NAME_SPACE": 2, "APP_ID": "55", "MESSAGE": None},
    },
    {
        "name": "ConTV",
        "country": ["usa", "can"],
        "id": "41",
        "config": {"NAME_SPACE": 2, "APP_ID": "18", "MESSAGE": None},
    },
    {
        "name": "Dove Channel",
        "country": ["usa", "can"],
        "id": "42",
        "config": {"NAME_SPACE": 2, "APP_ID": "16", "MESSAGE": None},
    },
    {
        "name": "Love Destination",
        "country": ["*"],
        "id": "64",
        "config": {"NAME_SPACE": 2, "APP_ID": "57", "MESSAGE": None},
    },
    {
        "name": "WatchFree",
        "country": ["usa"],
        "id": "48",
        "config": {"NAME_SPACE": 2, "APP_ID": "22", "MESSAGE": None},
    },
    {
        "name": "AsianCrush",
        "country": ["usa", "can"],
        "id": "50",
        "config": {
            "NAME_SPACE": 2,
            "APP_ID": "27",
            "MESSAGE": "https://html5.asiancrush.com/?ua=viziosmartcast",
        },
    },
]



"""Vizio SmartCast API base commands."""

from abc import abstractmethod
from typing import Any, Dict


class CommandBase(object):
    """Base command to send data to Vizio device."""

    def __init__(self, url: str = "") -> None:
        """Initialize base command to send data to Vizio device."""
        self._url = url

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__

    @property
    def _method(self) -> str:
        """Get command method."""
        return "put"

    @property
    def url(self) -> str:
        """Get endpoint for command."""
        return self._url

    @url.setter
    def url(self, new_url: str) -> None:
        """Set endpoint for command."""
        self._url = new_url

    def get_url(self) -> str:
        """Get endpoint for command."""
        return self._url

    def get_method(self) -> str:
        return self._method

    @abstractmethod
    def process_response(self, json_obj: Dict[str, Any]) -> None:
        """Always return True when there is no custom process_response method for subclass."""
        return True


class InfoCommandBase(CommandBase):
    """Base command to get data from Vizio device."""

    def __init__(self, url: str = "") -> None:
        """Initialize base command to get data from Vizio device."""
        super(InfoCommandBase, self).__init__(url)

    @property
    def _method(self) -> str:
        """Get command method."""
        return "get"

    @property
    def url(self) -> str:
        """Get endpoint for command."""
        return CommandBase.url.fget(self)

    @url.setter
    def url(self, new_url: str) -> None:
        """Set endpoint for command."""
        CommandBase.url.fset(self, new_url)

    def process_response(self, json_obj: Dict[str, Any]) -> bool:
        """Always return None when there is no custom process_response method for subclass."""
        return None



"""Vizio SmartCast API command and class for emulating remote key presses."""

from typing import List, Tuple

from pyvizio._api._protocol import ENDPOINT, KEY_ACTION
from pyvizio._api.base import CommandBase


class KeyPressEvent(object):
    """Emulated remote key press."""

    def __init__(
        self, key_code: Tuple[int, int], action: str = KEY_ACTION["PRESS"]
    ) -> None:
        """Initialize emulated remote key press."""
        self.CODESET = key_code[0]
        self.CODE = key_code[1]
        self.ACTION = action

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


class EmulateRemoteCommand(CommandBase):
    """Command to emulate remote key press."""

    def __init__(self, key_codes: List[Tuple[int, int]], device_type: str) -> None:
        """Initialize command to emulate remote key press."""
        super(EmulateRemoteCommand, self).__init__(ENDPOINT[device_type]["KEY_PRESS"])

        # noinspection SpellCheckingInspection
        self.KEYLIST = []

        for key_code in key_codes:
            self.KEYLIST.append(KeyPressEvent(key_code))



"""pyvizio helper functions."""

import asyncio
from functools import wraps
from typing import Any, Dict, List, Optional


def async_to_sync(f):
    """Decorator to run async function as sync."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def dict_get_case_insensitive(
    in_dict: Dict[str, Any], key: str, default_return: Any = None
) -> Any:
    """Case insensitive dict.get."""
    out_dict = {k.lower(): v for k, v in in_dict.items()}

    return out_dict.get(key.lower(), default_return)


def get_value_from_path(
    device_info: Dict[str, Any], paths: List[List[str]]
) -> Optional[Any]:
    """Iterate through known paths to return first valid path with value from Vizio device_info."""
    for path in paths:
        temp = ""
        for step in path:
            temp = dict_get_case_insensitive(device_info, step, {})

        if temp:
            return temp

    return None


# Adapted from https://gist.github.com/betrcode/0248f0fda894013382d7#gistcomment-3161499
async def open_port(host, port):
    """Return whether or not host's port is open.

    Parameters
    ----------
    host : str
        Host IP address or hostname
    port : int
        Port number

    Returns
    -------
    awaitable bool
    """
    try:
        _reader, writer = await asyncio.wait_for(
            asyncio.open_connection(host, port), timeout=2
        )
        writer.close()
        await writer.wait_closed()
        return True
    except Exception:
        pass

    return False



"""Vizio SmartCast API protocol constants and get and set functions."""

import json
from logging import Logger
from typing import Any, Dict

from aiohttp import ClientResponse, ClientSession, ClientTimeout
from aiohttp.client import DEFAULT_TIMEOUT as AIOHTTP_DEFAULT_TIMEOUT
import jsonpickle
from pyvizio._api.base import CommandBase
from pyvizio.const import DEVICE_CLASS_SPEAKER, DEVICE_CLASS_TV
from pyvizio.helpers import dict_get_case_insensitive

HTTP_OK = 200

ACTION_MODIFY = "MODIFY"

HEADER_AUTH = "AUTH"

STATUS_SUCCESS = "success"
STATUS_URI_NOT_FOUND = "uri_not_found"
STATUS_INVALID_PARAMETER = "invalid_parameter"

TYPE_EQ_SLIDER = "t_value_abs_v1"
TYPE_EQ_LIST = "t_list_v1"
TYPE_VALUE = "t_value_v1"

ENDPOINT = {
    DEVICE_CLASS_TV: {
        "BEGIN_PAIR": "/pairing/start",
        "FINISH_PAIR": "/pairing/pair",
        "CANCEL_PAIR": "/pairing/cancel",
        "INPUTS": "/menu_native/dynamic/tv_settings/devices/name_input",
        "CURRENT_INPUT": "/menu_native/dynamic/tv_settings/devices/current_input",
        "ESN": "/menu_native/dynamic/tv_settings/system/system_information/uli_information/esn",
        "SERIAL_NUMBER": "/menu_native/dynamic/tv_settings/system/system_information/tv_information/serial_number",
        "VERSION": "/menu_native/dynamic/tv_settings/system/system_information/tv_information/version",
        "DEVICE_INFO": "/state/device/deviceinfo",
        "POWER_MODE": "/state/device/power_mode",
        "KEY_PRESS": "/key_command/",
        "AUDIO_SETTINGS": "/menu_native/dynamic/tv_settings/audio",
        "CURRENT_APP": "/app/current",
        "LAUNCH_APP": "/app/launch",
    },
    DEVICE_CLASS_SPEAKER: {
        "BEGIN_PAIR": "/pairing/start",
        "FINISH_PAIR": "/pairing/pair",
        "CANCEL_PAIR": "/pairing/cancel",
        "INPUTS": "/menu_native/dynamic/audio_settings/input",
        "CURRENT_INPUT": "/menu_native/dynamic/audio_settings/input/current_input",
        "ESN": "/menu_native/dynamic/audio_settings/system/system_information/uli_information/esn",
        "SERIAL_NUMBER": "/menu_native/dynamic/audio_settings/system/system_information/speaker_information/serial_number",
        "VERSION": "/menu_native/dynamic/audio_settings/system/system_information/speaker_information/version",
        "DEVICE_INFO": "/state/device/deviceinfo",
        "POWER_MODE": "/state/device/power_mode",
        "KEY_PRESS": "/key_command/",
        "AUDIO_SETTINGS": "/menu_native/dynamic/audio_settings/audio",
    },
}

ITEM_CNAME = {
    "CURRENT_INPUT": "current_input",
    "ESN": "esn",
    "EQ": "eq",
    "POWER_MODE": "power_mode",
    "SERIAL_NUMBER": "serial_number",
    "VERSION": "version",
}

KEY_ACTION = {"DOWN": "KEYDOWN", "UP": "KEYUP", "PRESS": "KEYPRESS"}

KEY_CODE = {
    DEVICE_CLASS_TV: {
        "SEEK_FWD": (2, 0),
        "SEEK_BACK": (2, 1),
        "PAUSE": (2, 2),
        "PLAY": (2, 3),
        "DOWN": (3, 0),
        "LEFT": (3, 1),
        "OK": (3, 2),
        "UP": (3, 3),
        "LEFT2": (3, 4),
        "RIGHT": (3, 5),
        "BACK": (4, 0),
        "SMARTCAST": (4, 3),
        "CC_TOGGLE": (4, 4),
        "INFO": (4, 6),
        "MENU": (4, 8),
        "HOME": (4, 15),
        "VOL_DOWN": (5, 0),
        "VOL_UP": (5, 1),
        "MUTE_OFF": (5, 2),
        "MUTE_ON": (5, 3),
        "MUTE_TOGGLE": (5, 4),
        "PIC_MODE": (6, 0),
        "PIC_SIZE": (6, 2),
        "INPUT_NEXT": (7, 1),
        "CH_DOWN": (8, 0),
        "CH_UP": (8, 1),
        "CH_PREV": (8, 2),
        "EXIT": (9, 0),
        "POW_OFF": (11, 0),
        "POW_ON": (11, 1),
        "POW_TOGGLE": (11, 2),
    },
    DEVICE_CLASS_SPEAKER: {
        "PAUSE": (2, 2),
        "PLAY": (2, 3),
        "VOL_DOWN": (5, 0),
        "VOL_UP": (5, 1),
        "MUTE_OFF": (5, 2),
        "MUTE_ON": (5, 3),
        "MUTE_TOGGLE": (5, 4),
        "POW_OFF": (11, 0),
        "POW_ON": (11, 1),
        "POW_TOGGLE": (11, 2),
    },
}

PATH_MODEL = {
    DEVICE_CLASS_SPEAKER: [["name"]],
    DEVICE_CLASS_TV: [["model_name"], ["system_info", "model_name"]],
}


class PairingResponseKey(object):
    """Key names in responses to pairing commands."""

    AUTH_TOKEN = "auth_token"
    CHALLENGE_TYPE = "challenge_type"
    PAIRING_REQ_TOKEN = "pairing_req_token"


class ResponseKey(object):
    """Key names in responses to API commands."""

    HASHVAL = "hashval"
    CNAME = "cname"
    TYPE = "type"
    NAME = "name"
    VALUE = "value"
    METADATA = "metadata"
    ELEMENTS = "elements"
    ITEM = "item"
    ITEMS = "items"


async def async_validate_response(web_response: ClientResponse) -> Dict[str, Any]:
    """Validate response to API command is as expected and return response."""
    if HTTP_OK != web_response.status:
        raise Exception(
            "Device is unreachable? Status code: {0}".format(web_response.status)
        )

    try:
        data = json.loads(await web_response.text())
    except Exception:
        raise Exception("Failed to parse response: {0}".format(web_response.content))

    status_obj = dict_get_case_insensitive(data, "status")

    if not status_obj:
        raise Exception("Unknown response")

    result_status = dict_get_case_insensitive(status_obj, "result")

    if result_status and result_status.lower() == STATUS_INVALID_PARAMETER:
        raise Exception("invalid value specified")
    elif not result_status or result_status.lower() != STATUS_SUCCESS:
        raise Exception(
            "unexpected status {0}: {1}".format(
                result_status, dict_get_case_insensitive(status_obj, "detail")
            )
        )

    return data


async def async_invoke_api(
    ip: str,
    command: CommandBase,
    logger: Logger,
    custom_timeout: int = None,
    headers: Dict[str, Any] = {},
    log_api_exception: bool = True,
    session: ClientSession = None,
) -> Any:
    """Call API endpoints with appropriate request bodies and headers."""
    method = command.get_method()
    url = f"https://{ip}{command.get_url()}"
    data = jsonpickle.encode(command, unpicklable=False)

    if not custom_timeout:
        timeout = AIOHTTP_DEFAULT_TIMEOUT
    else:
        timeout = ClientTimeout(total=custom_timeout)

    if not headers:
        headers = {}

    try:
        if session:
            if "get" == method.lower():
                response = await session.get(
                    url=url, headers=headers, ssl=False, timeout=timeout
                )
            else:
                timeout = AIOHTTP_DEFAULT_TIMEOUT
                headers["Content-Type"] = "application/json"
                response = await session.put(
                    # url=url, data=str(data), headers=headers, ssl=False, timeout=timeout
                    url=url,
                    data=str(data),
                    headers=headers,
                    ssl=False,
                    timeout=timeout,
                )

            json_obj = await async_validate_response(response)
        else:
            async with ClientSession() as local_session:
                if "get" == method.lower():
                    response = await local_session.get(
                        url=url, headers=headers, ssl=False, timeout=timeout
                    )
                else:
                    timeout = AIOHTTP_DEFAULT_TIMEOUT
                    headers["Content-Type"] = "application/json"
                    response = await local_session.put(
                        url=url,
                        data=str(data),
                        headers=headers,
                        ssl=False,
                        timeout=timeout,
                    )

                json_obj = await async_validate_response(response)

        return command.process_response(json_obj)
    except Exception as e:
        if log_api_exception:
            logger.error("Failed to execute command: %s", e)

        return None


async def async_invoke_api_auth(
    ip: str,
    command: CommandBase,
    logger: Logger,
    auth_token: str = None,
    custom_timeout: int = None,
    log_api_exception: bool = True,
    session: ClientSession = None,
) -> Any:
    """Call auth protected API endpoints using CommandBase and subclass request bodies."""

    return await async_invoke_api(
        ip,
        command,
        logger,
        custom_timeout=custom_timeout,
        headers={HEADER_AUTH: auth_token},
        log_api_exception=log_api_exception,
        session=session,
    )



"""Vizio SmartCast API commands and class for pairing."""

from typing import Any, Dict

from pyvizio._api._protocol import ENDPOINT, PairingResponseKey, ResponseKey
from pyvizio._api.base import CommandBase
from pyvizio.helpers import dict_get_case_insensitive


class PairCommandBase(CommandBase):
    """Base pairing command."""

    def __init__(self, device_id: str, device_type: str, endpoint: str) -> None:
        """Initialize base pairing command."""
        super(PairCommandBase, self).__init__(ENDPOINT[device_type][endpoint])
        self.DEVICE_ID = device_id


class BeginPairResponse(object):
    """Response from command to begin pairing process."""

    def __init__(self, ch_type: str, token: str) -> None:
        """Initialize response from command to begin pairing process."""
        self.ch_type = ch_type
        self.token = token

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


class BeginPairCommand(PairCommandBase):
    """Command to begin pairing process."""

    def __init__(self, device_id: str, device_name: str, device_type: str) -> None:
        """Initialize command to begin pairing process."""
        super().__init__(device_id, device_type, "BEGIN_PAIR")
        self.DEVICE_NAME = str(device_name)

    def process_response(self, json_obj: Dict[str, Any]) -> BeginPairResponse:
        """Return response to command to begin pairing process."""
        item = dict_get_case_insensitive(json_obj, ResponseKey.ITEM)

        return BeginPairResponse(
            dict_get_case_insensitive(item, PairingResponseKey.CHALLENGE_TYPE),
            dict_get_case_insensitive(item, PairingResponseKey.PAIRING_REQ_TOKEN),
        )


class PairChallengeResponse(object):
    """Response from command to complete pairing process."""

    def __init__(self, auth_token: str) -> None:
        """Initialize response from command to complete pairing process."""
        self.auth_token = auth_token

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


class PairChallengeCommand(PairCommandBase):
    """Command to complete pairing process."""

    def __init__(
        self,
        device_id: str,
        challenge_type: int,
        pairing_token: int,
        pin: str,
        device_type: str,
    ) -> None:
        """Initialize command to complete pairing process."""
        super().__init__(device_id, device_type, "FINISH_PAIR")

        self.CHALLENGE_TYPE = int(challenge_type)
        self.PAIRING_REQ_TOKEN = int(pairing_token)
        self.RESPONSE_VALUE = str(pin)

    def process_response(self, json_obj: Dict[str, Any]) -> PairChallengeResponse:
        """Return response to command to complete pairing process."""
        item = dict_get_case_insensitive(json_obj, ResponseKey.ITEM)

        return PairChallengeResponse(
            dict_get_case_insensitive(item, PairingResponseKey.AUTH_TOKEN)
        )


class CancelPairCommand(PairCommandBase):
    """Command to cancel pairing process."""

    def __init__(self, device_id, device_name: str, device_type: str) -> None:
        """Initialize command to cancel pairing process."""
        super().__init__(device_id, device_type, "CANCEL_PAIR")

        self.DEVICE_NAME = str(device_name)
